function dlb = lowbo(A,b,x,h,mode)
% =========================================================================
%  FILE:           lowbo.m
%  AUTHOR:         Andreas Tillmann, TU Darmstadt, Germany
%  LAST MODIFIED:  06/21/2011
% =========================================================================
%  DESCRIPTION:    This function computes a valid lower bound for the
%                  objective function value of the L1-norm minimization
%                  problem (P1):  min ||x||_1  s.t.  Ax=b (underdet. sys.)
%                  by generating a feasible dual point y, i.e., y s.t.
%                  ||A'*y||_inf <= 1, and evaluating the dual ob-
%                  jective function -b'y. Several constructions of y are
%                  possible to choose from; see description of 'mode'.
%
%  INPUT:          A    The constraint system matrix of the (P1) instance
%                  b    The right hand side vector of the constraints
%                  x    A vector interpreted as a primal solution; may be
%                       an arbitrary (infeas.) vector (of correct length)
%                  h    A (stabilized) subgradient of ||.||_1 at x
%                  mode A parameter specifying how the dual lower bound is
%                       computed. Feasible values for mode are:
%                       0 = Constant lower bound of zero
%                       1 = A simple constant lower bound is calculated 
%                           from b, based on the following idea:
%                           Unless b=0, ||x||>0 for all x solving Ax=b, 
%                           and considering the dual objective  max{-b'y},
%                           one can obtain a valid positive bound by using
%                           the projection of -b onto the dual feasible set
%                       2 = A valid dual lower bound is obtained by (appro-
%                           ximately) solving the system 
%                                       A*A'*y = -A*sign(x)
%                           and then projecting y onto the dual constraint 
%                           set. This approach is based on following ideas:
%                           Consider the Lagrange function to (P1),
%                                L(x,y) = ||x||_1 + y'(Ax-b),
%                           which can be viewed as a kind of penalty func-
%                           tion. Wishing to minimize L(x,y) (instead of 
%                           ||x||_1 with explicit constraints Ax=b), we
%                           consider one subgradient of L(x,y) in x, namely
%                           sign(x)+A'*y. A minimum would have zero in the 
%                           corresponding subdifferential, thus we try to 
%                           solve sign(x)+A'*y=0 for y (x is given here). 
%                           When Ax=b is underdetermined, this system is 
%                           overdetermined and may not have a solution at
%                           all; we therefore solve the normal equation
%                           A*A'*y = -A*sign(x) instead. The Lagrange mul-
%                           tiplicators y are, in a sense, already "dual" 
%                           variables, but we need to ensure dual feasibi-
%                           lity to obtain a valid dual lower bound -b'*y.
%                           Hence, the projection of y onto the dual fea-
%                           sible set is used.
%                           This idea is related to the Heuristic Optimality
%                           Check routine; see HOC.m in L1TestPack,
%                           available at wwwopt.mathematik.tu-darmstadt.de/spear
%
%                       3 = A variation of 2, sometimes yielding better
%                           bounds: Instead of the subgradient h=sign(x),
%                           only the sign of the x-part corresp. to entries
%                           with abs.val. >=1 is used, while for the remai-
%                           ning entries y(i) the values x(i)<1 are used.
%                       4 = Another variation of 2 using a (possibly 
%                           stabilized) subgradient of x (user-supplied as
%                           argument h) instead of sign(x).
%                       5 = A variation of 2 which uses a Camerini/Mattioli/
%                           Fratta-type stabilized subgradient. When used
%                           to estimate the duality gap in an algorithm, the
%                           function argument h should be the CMF-type stab. 
%                           subgradradient of the previous iterate; 
%                           seems to yield best overall duality gap esti-
%                           mates (considering trade-off with comp. time)
%                       6 = A variation of 2, even more related to the HOC:
%                           Instead of taking the full vector h, only h-
%                           entries that correspond to the support of the 
%                           thresholded vector x are used, i.e., entries of
%                           x with absolute values smaller than 1e-9 are
%                           treated as zeros. The dual variable is then 
%                           computed using only the reduced-dimensional
%                           system A(:,S)'*y=-h(S), where S is the
%                           support of the thresholded x-vector. Scaling
%                           y to be dual feasible yields the lower bound
%                           -b'*y. Compare with mode=2 and HOC.m.
%                           NOTE: uses function supp.m
%                       [ optional argument; default is mode=5 ]
%
%  OUTPUT:         dlb  The (dual) lower bound on the (primal) objective
%                       function value, i.e., dlb <= ||x||_1 for x: Ax=b
% =========================================================================

if( nargin~=5 )
    mode = 5;
end

switch mode

    case 0, % 0 is always a feasible lower bound for ||x||_1;
            % x,h,A,b not needed (can be empty)
        dlb = 0;

    case 1, % Simple dual lower bound; x, h not needed (can be empty).
        y = -b./norm(-A'*b,'inf'); 
        dlb = -b'*y;

    case 2, % h not needed (sign is used, h can be empty)
        y = CG( A, [], -A*sign(x), 1e-5, 3, 1 );
        y = y./norm(A'*y,'inf');
        dlb = -b'*y;

    case 3, % variation of case 2: if |x(i)|<1, init. ylb(i)=x(i);
            % h not needed (can be empty)
        idx = find(abs(x)<1); s = sign(x); s(idx) = x(idx);
        y = CG( A, [], -A*s, 1e-5, 3, 1 );
        y = y./norm(A'*y,'inf');
        dlb = -b'*y;

    case 4, % variation of case 2: h is user-provided subgradient of x
        y = CG( A, [], -A*h, 1e-5, 3, 1 );        
        y = y./norm(A'*y,'inf');
        dlb = -b'*y;

    case 5, % variation of 2 inspired by CMF-style subgradient 
            % stabilization: h should be stab. subgradient of the
            % iterate previous to x when used in an algorithm
            % (h could be just any subgradient, as well)
        s = sign(x);
        s = s + norm(s)/norm(h)*h;
        y = CG( A, [], -A*s, 1e-5, 3, 1 );
        y = y./norm(A'*y,'inf');
        dlb = -b'*y;
        % Seems to provide best bounds in average, therefore set as default

    case 6, % variation of 2 using approximate supports as in HOC
            % h not needed (can be empty)
         [Sx,~] = supp(x,1e-9);
         ASx = A(:,Sx);
         w = CG( ASx, [], -ASx*h(Sx), 1e-9, 20, 1 );
         y = w./norm(A'*w,'inf');
         dlb = -b'*y;

    otherwise,
        disp('Invalid mode. Choose mode between 0 and 6.\n');
        dlb = [];
end

% negative bounds are worse than trivial bounds (case 0 or 1);
% in this case, we override dlb by 0:
dlb = max(dlb,0);
% dlb = max(dlb,b'*b/norm(-A'*b,'inf'));  % could also use case 1

if( isnan(dlb) ) % this can happen due to numerical instability problems
    dlb = 0;
    disp('Computation of dual lower bound returned NaN. dLB set to zero.');
end

if( issparse(dlb) )  % fprintf will not work with sparse input
    dlb = full(dlb);
end
